#include "RigidBody.h"

/**
*	Constructor for a RigidBody object.
*	Sets the shape of the RigidBody and the maxVerts and maxEdges based on the shape.
*	Does not create Vertices or Edges for the shape. This can be done with the finishRigidBody method.
*/
RigidBody::RigidBody(Shape s)
{
	shape = s;
	vertexCount = 0;
	edgeCount = 0;

	switch(shape)
	{
	case OBSTACLE:
		maxVerts = 4;
		maxEdges = 6;
		break;
	case TRIANGLE:
		maxVerts = 3;
		maxEdges = 3;
		break;
	case SQUARE:
		maxVerts = 4;
		maxEdges = 6;
		break;
	default:
		cout << "Error: Cannot create RigidBody of type " << shape << endl;
		exit(1);
		break;
	}
}

RigidBody::~RigidBody()
{
	for(int i = 0; i < vertexCount; ++i)
		delete vertexArray[i];

	for(int i = 0; i < edgeCount; ++i)
		delete edgeArray[i];
}

/**
*	Creates the edges and vertices for the RigidBody based on the shape of the RigidBody.
*	Also sets the positions of the edges and vertices based on x and y position values.
*	x: x coordinate of RigidBody in world
*	y: y coordinate of RigidBody in world
*/
void RigidBody::finishRigidBody(int x, int y)
{
	switch(shape)
	{
	case OBSTACLE:
		{
			Vertex* v1 = new Vertex(x, y);
			Vertex* v2 = new Vertex(x + OBSTACLE_WIDTH, y);
			Vertex* v3 = new Vertex(x + OBSTACLE_WIDTH, y + OBSTACLE_HEIGHT);
			Vertex* v4 = new Vertex(x, y + OBSTACLE_HEIGHT);

			vertexArray[0] = v1;
			vertexArray[1] = v2;
			vertexArray[2] = v3;
			vertexArray[3] = v4;

			//create border edges
			edgeArray[0] = new Edge(v1, v2, true);
			edgeArray[1] = new Edge(v2, v3, true);
			edgeArray[2] = new Edge(v3, v4, true);
			edgeArray[3] = new Edge(v4, v1, true);

			//create cross section edges (non-border)
			edgeArray[4] = new Edge(v1, v3, false);
			edgeArray[5] = new Edge(v2, v4, false);

			vertexCount = 4;
			edgeCount = 6;
			break;
		}
	case TRIANGLE:
		{
			Vertex* v1 = new Vertex(x, y);
			Vertex* v2 = new Vertex(x - 25, y - 25);
			Vertex* v3 = new Vertex(x - 50, y);
	
			vertexArray[0] = v1;
			vertexArray[1] = v2;
			vertexArray[2] = v3;

			edgeArray[0] = new Edge(v1, v2);
			edgeArray[1] = new Edge(v2, v3);
			edgeArray[2] = new Edge(v3, v1);

			vertexCount = 3;
			edgeCount = 3;
		}
		break;
	case SQUARE:
		{
			Vertex* v1 = new Vertex(x, y);
			Vertex* v2 = new Vertex(x + SQUARE_LENGTH, y);
			Vertex* v3 = new Vertex(x + SQUARE_LENGTH, y + SQUARE_LENGTH);
			Vertex* v4 = new Vertex(x, y + SQUARE_LENGTH);

			vertexArray[0] = v1;
			vertexArray[1] = v2;
			vertexArray[2] = v3;
			vertexArray[3] = v4;

			//create border edges
			edgeArray[0] = new Edge(v1, v2, true);
			edgeArray[1] = new Edge(v2, v3, true);
			edgeArray[2] = new Edge(v3, v4, true);
			edgeArray[3] = new Edge(v4, v1, true);

			//create cross section edges (non-border)
			edgeArray[4] = new Edge(v1, v3, false);
			edgeArray[5] = new Edge(v2, v4, false);

			vertexCount = 4;
			edgeCount = 6;
		}
		break;
	default:
		{
			cout << "Error: Cannot create RigidBody of type " << shape << endl;
			exit(1);
		}
		break;//should never reach
	}
}

/**
*	Adds the specified edge to the RigidBody. Increases the edge count.
*	e: Edge to add
*/
void RigidBody::AddEdge(Edge* e) {
	if(edgeCount + 1 <= maxEdges) {
		edgeArray[edgeCount] = e;
		edgeCount++;
	}
}

/**
*	Adds the specified Vertex to the RigidBody. Increases the vertex count.
*	v: Vertex to add
*/
void RigidBody::AddVertex(Vertex *v) {
	if(vertexCount + 1 <= maxVerts) {
		vertexArray[vertexCount] = v;
		vertexCount++;
	}
}

/**
*	Projects the Rigid body to the specified axis. Then stores the min and max
*	values calculated in the specified min and max variables.
*/
void RigidBody::ProjectToAxis(Vec2& axis, float& min, float& max) {
	float DotProduct = axis * vertexArray[0]->pos;

	//Set the min and max values based on the new projection

	//start by setting to projection of first vert
	min = DotProduct;
	max = DotProduct;

	//then iterate through remaining verts and update if necessary
	for(int i = 0; i < vertexCount; i++) {
		DotProduct = axis * vertexArray[i]->pos;

		min = min(DotProduct, min);
		max = max(DotProduct, max);
	}
}

/**
*	Determines the center of the RigidBody and sets the center field.
*/
void RigidBody::CalculateCenter() 
{
	center = Vec2(0.0f, 0.0f);

	//initialize min and max values to extreme values
	minX = minY =  10000.0f;
	maxX = maxY = -10000.0f;

	//loop through all vertices in the array
	for(int i = 0; i < vertexCount; i++)
	{
		//add the pos of the vertex to the center
		center += vertexArray[i]->pos;

		//adjust the min and max x,y values
		minX = min(minX, vertexArray[i]->pos.x);
		minY = min(minY, vertexArray[i]->pos.y);
		maxX = max(maxX, vertexArray[i]->pos.x);
		maxY = max(maxY, vertexArray[i]->pos.y);
	}

	//divide by the number of vertices to find the average, which is the center
	center = center / vertexCount;
}