#include "World.h"

//initialize instance to nothing
World* World::instance = NULL;

/**
*	Returns the Singleton instance of the World
*/
World* World::getInstance()
{
	if(!instance)
		instance = new World();

	return instance;
}

/**
*	Constructor for a World object
*	Sets gravity to the bottom of the screen
*	No bodies, edges, or vertices in the world
*/
World::World()
{
	gravity = Vec2(0.0f, 0.2f);
	totalObstacles = 0;
	totalBodies = 0;
	totalVertices = 0;
	totalEdges = 0;

	tick = TICK;
	iterations = 12;//number of times to re-run the updateEdges function per tick
}

void World::Destroy()
{
	for(int i = 0; i < totalBodies; ++i)
		delete RigidBodyArray[i];

	for(int i = 0; i < totalObstacles; ++i)
		delete ObstacleArray[i];
}

int World::getIterations()
{
	return iterations;
}

void World::increaseIterations()
{
	if(iterations < MAX_ITERATIONS)
	{
		iterations++;
	}
}

void World::decreaseIterations()
{
	if(iterations > 1)
	{
		iterations--;
	}
}

void World::setGravity(Vec2 newGravity)
{
	gravity = newGravity;
}

/**
*	Add a RigidBody to the World
*	body1:	RigidBody object to add
*/
void World::AddBody(RigidBody* body1) 
{
	if(totalBodies + 1 < MAX_OBJECTS)
	{
		RigidBodyArray[totalBodies++] = body1;
	}
}

/**
*	Add an Obstacle to the World
*	body1: RigidBody obstacle to add
*/
void World::AddObstacle(RigidBody* obstacle)
{
	if(totalObstacles + 1 < MAX_OBSTACLES)
	{
		obstacle->CalculateCenter();
		ObstacleArray[totalObstacles++] = obstacle;
	}
}

/**
*	Updates each Vertex in the World based on the forces in the World
*	Currently, the only force applied is gravity
*/
void World::updateForces() 
{
	for(int i = 0; i < totalBodies; i++)
	{
		RigidBody* rb = RigidBodyArray[i];
		for(int j = 0; j < rb->vertexCount; j++)
		{
			rb->vertexArray[j]->accel = gravity;
		}
	}
}

/**
*	Updates the position of each Vertex in the World
*/
void World::updateVerlet() 
{
	for(int i = 0; i < totalBodies; i++)
	{
		RigidBody* rb = RigidBodyArray[i];

		for(int j = 0; j < rb->vertexCount; j++)
		{
			Vertex& v = *(rb->vertexArray[j]);

			//xNew = 2 * x(Current)  x(Old) + a(Current) * t2
			//wikipedia.org/wiki/Verlet_integration

			Vec2 temp = v.pos;
			v.pos = (v.pos * 2) - v.oldPos + v.accel * tick * tick;
			v.oldPos = temp;
		}
	}
}

/**
*	Updates the length of every Edge in the World.
*	Each update causes the length of the edge to change.
*	This method adjusts the positions of the vertices of an edge to maintain
*	the original length.
*/
void World::updateEdges() 
{
	//loop through all possible edges
	for(int i = 0; i < totalBodies; i++)
	{
		RigidBody* rb = RigidBodyArray[i];

		for(int j = 0; j < rb->edgeCount; j++)
		{
			Edge& e = *(rb->edgeArray[j]);

			Vec2 edgeVector = e.vertex2->pos - e.vertex1->pos;

			//move the positions of both vertices so the length of the edge is maintained
			float edgeVectorLength = edgeVector.length();
			float d = edgeVectorLength - e.length;
		
			edgeVector.normalize();

			//shift each vertex over half the total while maintaining length
			e.vertex1->pos += edgeVector * d * 0.5f;
			e.vertex2->pos -= edgeVector * d * 0.5f;
		}
	}
}

/**
*	Iterates through every RigidBody in the World, updates their edges, keeps them inside the screen,
*	and handles any collisions that occur.
*/
void World::processAllCollisions() {
	//the more times we iterate through all the edges and adjust them, the more accurate the result is
	//use the iterations variable to determine how many times to iterate
	for(int i = 0; i < iterations; i++) 
	{
		//if vertices are outside of screen, move them back inside
		for(int j = 0; j < totalBodies; j++)
		{
			RigidBody* rb = RigidBodyArray[j];
			for(int k = 0; k < rb->vertexCount; k++)
			{
				Vec2& position = rb->vertexArray[k]->pos;
	
				position.x = max(min(position.x, (float)SCREEN_WIDTH), 0.0f);
				position.y = max(min(position.y, (float)SCREEN_HEIGHT), 0.0f);
			}
		}

		updateEdges();

		for(int k = 0; k < totalBodies; k++)
		{
			RigidBodyArray[k]->CalculateCenter();
		}

		//iterate through every body in the word
		for(int body1 = 0; body1 < totalBodies; body1++) 
		{
			for(int body2 = 0; body2 < totalBodies; body2++) 
			{
				if(body1 != body2)
				{
					if(didCollide(RigidBodyArray[body1], RigidBodyArray[body2]))
					{
						handleCollision();
					}
				}
			}

			//check against obstacles
			for(int i = 0; i < totalObstacles; i++)
			{
				if(didCollide(RigidBodyArray[body1], ObstacleArray[i]))
				{
					handleCollision();
				}
			}
		}
	}
}

/**
*	Takes two RigidBodies and determines whether they are overlapping.
*	body1:	First body to check for overlap
*	body2:	Second body to check for overlap
*	return:	true if RigidBodies are overlapping
*			false if RigidBodies are not overlapping
*/
bool World::overlapping(RigidBody* body1, RigidBody* body2)
{

	if ((body1->minX > (body2->maxX)) || ((body1->maxX) < body2->minX))
		return false;
	
	if((body1->minY > (body2->maxY)) || ((body1->maxY) < body2->minY))
		return false;

	return true;
}

/**
*	Takes two RigidBodies and determines whether the two collided
*	body1:	First RigidBody to check for collision
*	body2:	Second RigidBody to check for collision
*	return:	true if RigidBodies are colliding
*			false if RigidBodies are not colliding
*/
bool World::didCollide(RigidBody* body1, RigidBody* body2)
{
	//if they aren't overlapping, they didn't collide
	if(!overlapping(body1, body2))
		return false;

	//set the minDist to an extreme value
	float minDist = 10000.0f;

	//iterate through the edges of both bodies
	for(int i = 0; i < body1->edgeCount + body2->edgeCount; i++)
	{
		Edge* e;

		if(i < body1->edgeCount)
		{
			e = body1->edgeArray[i];
			currentImpact.edgeParent = body1;
		}
		else
		{
			e = body2->edgeArray[i - body1->edgeCount];
			currentImpact.edgeParent = body2;
		}

		//This will skip edges that lie totally inside the bodies, as they don't matter.
		if(e->isBorder)
		{
			Vec2 axis(e->vertex1->pos.y - e->vertex2->pos.y,
					  e->vertex2->pos.x - e->vertex1->pos.x);
			axis.normalize();

			//project the two bodies
			float minA, minB, maxA, maxB;
			body1->ProjectToAxis(axis, minA, maxA);
			body2->ProjectToAxis(axis, minB, maxB);

			float dist = getDistance(minA, maxA, minB, maxB);

			if(dist > 0.0f)
				return false;//no overlap = no collision
			else if(abs(dist) < minDist) {
				minDist = abs(dist);

				currentImpact.normal = axis;
				currentImpact.edge = e;
			}
		}
	}

	currentImpact.depth = minDist;

	if(currentImpact.edgeParent != body2) 
	{
		RigidBody* temp = body2;
		body2 = body1;
		body1 = temp;
	}

	int sign = sgn(currentImpact.normal * (body1->center - body2->center));

	if(sign != 1)
		currentImpact.normal = -currentImpact.normal;//if the normal points away from the other body, switch it

	Vec2 CollisionVector = currentImpact.normal * currentImpact.depth;

	//set minDist2 to an extreme value
	float minDist2 = 10000.0f;
	for(int j = 0; j < body1->vertexCount; j++)
	{
		//get distance of vertex to line
		float dist2 = currentImpact.normal * (body1->vertexArray[j]->pos);

		if(dist2 < minDist2)
		{
			minDist2 = dist2;
			currentImpact.vertex = body1->vertexArray[j];
		}
	}

	//objects collided
	return true;
}

/**
*	Called after a collision between a vertex and an edge has been detected and both structs have been
*	stored in currentImpact. handleCollision calculates a vector which will push the vertex and the edge apart
*	so they are no longer touching. If the vertex or edge belongs to a RigidBody of type OBSTACLE, it will not be moved.
*	Once the vector has been calculated, it will be applied to the appropriate vertex and/or edge.
*/
void World::handleCollision() 
{
	Vertex* vertex1 = currentImpact.edge->vertex1;
	Vertex* vertex2 = currentImpact.edge->vertex2;

	Vec2 CollisionVector = currentImpact.normal * currentImpact.depth;


	//the vertex can hit at any point along the vertex, find out where it hit
	float edgePos;
	if(abs(vertex1->pos.x - vertex2->pos.x ) > abs(vertex1->pos.y - vertex2->pos.y))
	{//check for divide by 0 error
		edgePos = (currentImpact.vertex->pos.x - CollisionVector.x - vertex1->pos.x) / (vertex2->pos.x - vertex1->pos.x);
	}
	else
	{
		edgePos = (currentImpact.vertex->pos.y - CollisionVector.y - vertex1->pos.y) / (vertex2->pos.y - vertex1->pos.y);
	}

	//calculate an offset scale so the vertex on the edge closest to the collision vertex is most strongly affected
	float scale = 1.0f/(edgePos*edgePos + (1 - edgePos) * (1 - edgePos));

	//apply vectors for impact
	if(currentImpact.edgeParent->shape != OBSTACLE)
	{
		vertex1->pos -= CollisionVector * (1 - edgePos) * 0.5f * scale;
		vertex2->pos -= CollisionVector * edgePos * 0.5f * scale;
	}
	
	currentImpact.vertex->pos += CollisionVector * 0.5f;
}

/**
*	Returns the distance between a Body at minA and maxA and a Body at minB and maxB
*/
float World::getDistance(float minA, float maxA, float minB, float maxB)
{
	if(minA < minB)
		return minB - maxA;
	else
		return minA - maxB;
}

/**
*	The update loop for the world. For each RigidBody in the world, applies
*	gravity, updates positions, and handles collision detection and collision
*	responses.
*/
void World::Update() 
{
	updateForces();
	updateVerlet();
	processAllCollisions();
}

/**
*	Draws each RigidBody present in the World
*	Squares are drawn in blue, triangles in green
*	Call each update to redraw the scene
*/
void World::Render() {
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_LINE_SMOOTH);

	//draw quads and tris for each shape
	for(int j = 0; j < totalBodies; j++)
	{
			switch(RigidBodyArray[j]->shape)
			{
			case TRIANGLE:
				glColor3f(0.53f, 0.8f, 0.53f);
				glBegin(GL_TRIANGLES);
					glVertex2f(RigidBodyArray[j]->vertexArray[0]->pos.x, RigidBodyArray[j]->vertexArray[0]->pos.y);
					glVertex2f(RigidBodyArray[j]->vertexArray[1]->pos.x, RigidBodyArray[j]->vertexArray[1]->pos.y);
					glVertex2f(RigidBodyArray[j]->vertexArray[2]->pos.x, RigidBodyArray[j]->vertexArray[2]->pos.y);
				glEnd();
				break;
			case SQUARE:
				glColor3f(0.53f, 0.8f, 1.0f);
				glBegin(GL_QUADS);
					glVertex2f(RigidBodyArray[j]->vertexArray[0]->pos.x, RigidBodyArray[j]->vertexArray[0]->pos.y);
					glVertex2f(RigidBodyArray[j]->vertexArray[1]->pos.x, RigidBodyArray[j]->vertexArray[1]->pos.y);
					glVertex2f(RigidBodyArray[j]->vertexArray[2]->pos.x, RigidBodyArray[j]->vertexArray[2]->pos.y);
					glVertex2f(RigidBodyArray[j]->vertexArray[3]->pos.x, RigidBodyArray[j]->vertexArray[3]->pos.y);
				glEnd();
				break;
			default:
				break;
			}
	}

	//draw outlines of each shape
	glLineWidth(2.0f);
	for(int i = 0; i < totalBodies; i++)
	{
		RigidBody* rb = RigidBodyArray[i];

		switch(rb->shape)
		{
		case TRIANGLE:
			glColor3f(0.0f, 0.55f, 0.0f);//light green
			break;
		case SQUARE:
			glColor3f(0.0f, 0.5f, 1.0f);//light blue
			break;
		default:
			glColor3f(0.0f, 0.5f, 1.0f);//same as square, light blue
			break;
		}

		for(int j = 0; j < rb->edgeCount; j++)
		{
			glBegin(GL_LINES);
			if(rb->edgeArray[j]->isBorder)
			{
				glVertex2f(rb->edgeArray[j]->vertex1->pos.x, rb->edgeArray[j]->vertex1->pos.y);
				glVertex2f(rb->edgeArray[j]->vertex2->pos.x, rb->edgeArray[j]->vertex2->pos.y);
			}
		glEnd();
		}
	}

	//draw Obstacles in Obstacle Array
	for(int i = 0; i < totalObstacles; i++)
	{
		switch(ObstacleArray[i]->shape)
		{
		case OBSTACLE:
				glColor3f(0.52f, 0.37f, 0.26f);//light brown
				glBegin(GL_QUADS);
					glVertex2f(ObstacleArray[i]->vertexArray[0]->pos.x, ObstacleArray[i]->vertexArray[0]->pos.y);
					glVertex2f(ObstacleArray[i]->vertexArray[1]->pos.x, ObstacleArray[i]->vertexArray[1]->pos.y);
					glVertex2f(ObstacleArray[i]->vertexArray[2]->pos.x, ObstacleArray[i]->vertexArray[2]->pos.y);
					glVertex2f(ObstacleArray[i]->vertexArray[3]->pos.x, ObstacleArray[i]->vertexArray[3]->pos.y);
				glEnd();
			break;
		default:
			break;
		}
	}

	//print obstacle edges
	for(int i = 0; i < totalObstacles; i++)
	{
		RigidBody* rb = ObstacleArray[i];

		switch(rb->shape)
		{
		case OBSTACLE:
			glColor3f(0.65f, 0.5f, 0.39f);//light brown
			break;
		default:
			glColor3f(1.0f, 1.0f, 1.0f);//white
			break;
		}

		for(int j = 0; j < rb->edgeCount; j++)
		{
			glBegin(GL_LINES);
			if(rb->edgeArray[j]->isBorder)
			{
				glVertex2f(rb->edgeArray[j]->vertex1->pos.x, rb->edgeArray[j]->vertex1->pos.y);
				glVertex2f(rb->edgeArray[j]->vertex2->pos.x, rb->edgeArray[j]->vertex2->pos.y);
			}
		glEnd();
		}
	}
}